#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zlib.h"

#pragma pack(1)
struct resinfostruct {
  int nameofs;
  int ofs;
  int len;
  int decompressed_len;
  int algo;
  int csum;
};
struct cflheaderstruct {
  int ofs;
  int len;
  int decompressed_len;
  int algo;
  int csum;
};
#pragma pack()

#define MIN(a,b) (((a)<(b))?(a):(b))
#define CHECKSUM_ASCIIZ -1
#define HEADERMAGIC 0x324C4643
#define FOOTERMAGIC 0x4C464332

char * CFL_strdup(char * asciiz) {
  char * buf;
  int i;
  i=strlen(asciiz);
  buf=malloc(i+1);
  memset(buf,0,i+1);
  memcpy(buf,asciiz,i);
  return buf;
}

int CFL_preprocess(char * buf, int blen, int algo) {
  char * temp;
  if (algo==0)
    return 0;
  temp=malloc(blen);
  memcpy(temp,buf,blen);
  /* temp -> buf tnx */
  switch (algo) {
  default: printf("\n\nPreprocess: unknown algorithm\n");
           exit(1);
           break;
  }
  free(temp);
  return 0;
}

int CFL_postprocess(char * buf, int blen, int algo) {
  char * temp;
  if (algo==0)
    return 0;
  temp=malloc(blen);
  memcpy(temp,buf,blen);
  /* temp -> buf tnx */
  switch (algo) {
  default: printf("\n\nPostprocess: unknown algorithm\n");
           exit(1);
           break;
  }
  free(temp);
  return 0;
}

char * CFL_compress(char * src, int srclen, int * destlen, int algo) {
  char * outbuf;
  outbuf=malloc(*destlen);
  switch (algo) {
  case 0: memcpy(outbuf,src,MIN(srclen,*destlen)); *destlen=MIN(srclen,*destlen); break;
  case 1: if (compress(outbuf,(unsigned long*)destlen,src,srclen)!=Z_OK) {
            free(outbuf);
            outbuf=NULL;
          }
          break;
  case 2: if (compress(outbuf,(unsigned long*)destlen,src,srclen)!=Z_OK) {
            free(outbuf);
            outbuf=NULL;
          }
          break;
  default: printf("\n\nCompress: unknown algorithm\n");
           exit(1);
           break;
  }
  return outbuf;
}

#ifdef __WATCOMC__
unsigned int ROL(unsigned int i);
#pragma aux ROL="rol eax,3" parm [EAX] value [EAX] modify nomemory;
#endif

/*
 * Slightly different checksum, giving different values to
 * ABBA and BABA. Not quite CRC but certainly faster =)
 */
unsigned int CFL_checksum(char * buf, int buflen) {
  unsigned int a,i=0;
#ifndef __WATCOMC__
  unsigned int q;
#endif
  if (buflen==CHECKSUM_ASCIIZ) { /* check ASCIIZ string instead */
    while (*buf) {
      i+=*buf;
#ifdef __WATCOMC__
      i=ROL(i);
#else
      q=i>>29; /* 3 high bits thankyou. */
      i<<=3;   /* practically this is ROL EAX,3 */
      i+=q;    /* instruction. C high level? yea rite! =) */
#endif
      buf++;
    }
  } else {
    for (a=0;a<buflen;a++) {
      i+=*(buf+a);
#ifdef __WATCOMC__
      i=ROL(i);
#else
      q=i>>29;
      i<<=3;
      i+=q;
#endif
    }
  }
  return i;
}

char * CFL_pack(char * src, int srclen, int * destlen, int algo, int * csum) {
  char * outbuf;
  *csum=CFL_checksum(src,srclen);
  if (CFL_preprocess(src,srclen,(algo/100)%100)) {
    free(src);
    printf("\n\nPreprocess: failed.\n");
    exit(1);
  }
  *destlen=srclen*2;
  if ((outbuf=CFL_compress(src,srclen,destlen,algo%100))==NULL) {
    free(src);
    printf("\n\nCompress: failed.\n");
  }
  if (CFL_postprocess(outbuf,*destlen,algo/10000)) {
    free(src);
    free(outbuf);
    printf("\n\nPostprocess: failed.\n");
  }
  free(src);
  return outbuf;
}

int CFL_dosexists(char * fname) {
  FILE * f;
  f=fopen(fname,"rb");
  if (f==NULL)
    return 0;
  fclose(f);
  return 1;
}

int filelen(char * fname) {
  FILE * f;
  int i;
  f=fopen(fname,"rb");
  if (f==NULL)
    return 0;
  fseek(f,0,SEEK_END);
  i=ftell(f);
  fclose(f);
  return i;
}

char * getfile(char * fname) {
  FILE * f;
  char * buf;
  int i;
  i=filelen(fname);
  f=fopen(fname,"rb");
  if (f==NULL)
    return NULL;
  buf=malloc(i);
  fread(buf,i,1,f);
  fclose(f);
  return buf;
}

void read_line(FILE * f, char * buf) {
  int i,fine=0;
  memset(buf,0,255);
  do {
    i=fgetc(f);
    switch (i) {
    case 13:break;
    case EOF:
    case 10:fine=1; break;
    default:*buf=i; buf++;
    }
  } while (!fine);
}

char * targetfile=NULL;
char * infostring="Compressed File Library version 2.0";
char * checkstring="CFL 2.0";
char * outbuf;
FILE * outf;
int libpack=3;
int resources=0;

void ini_pass1(void) {
  FILE * f;
  char * buf;
  int i,a;
  buf=malloc(256);
  f=fopen("cfl2.ini","rb");
  resources=0;
  do {
    read_line(f,buf);
    switch (*buf) {
    case ';':
    case  0 : break;
    case '#': if (strnicmp(buf+1,"target=",7)==0) {
                i=0;
                a=0;
                while (*(buf+a)!='\"') a++;
                a++;
                while (*(buf+a+i)!='\"') i++;
                *(buf+a+i)=0;
                targetfile=strdup(buf+a);
              } else
              if (strnicmp(buf+1,"info=",5)==0) {
                i=0;
                a=0;
                while (*(buf+a)!='\"') a++;
                a++;
                while (*(buf+a+i)!='\"') i++;
                *(buf+a+i)=0;
                infostring=strdup(buf+a);
              } else
              if (strnicmp(buf+1,"check=",6)==0) {
                i=0;
                a=0;
                while (*(buf+a)!='\"') a++;
                a++;
                while (*(buf+a+i)!='\"') i++;
                *(buf+a+i)=0;
                checkstring=strdup(buf+a);
              } else
              if (strnicmp(buf+1,"libpack=",8)==0) {
                libpack=strtol(buf+9,0,0);
              } else {
                printf("pass1: unknown config line:\n\t\"%s\"\n",buf);
                exit(1);
              }
              break;
    default : resources++; break;
    }
  } while (!feof(f));
  free(buf);
  fclose(f);
}

int pass2(void) {
  FILE * f;
  char * buf, * dosfname;
  int i,a,check=0;
  dosfname=malloc(256);
  buf=malloc(256);
  f=fopen("cfl2.ini","rb");
  do {
    memset(dosfname,0,255);
    read_line(f,buf);
    switch (*buf) {
    case ';':
    case  0 :
    case '#': break;
    default : i=0;
              while ((*(buf+i)!='=') && (*(buf+i)!=0)) {
                *(dosfname+i)=*(buf+i);
                i++;
              }
              *(dosfname+i)=0;
              if (!CFL_dosexists(dosfname)) {
                printf("Error: \"%s\" ",dosfname);
                printf("Doesn't exist!\n");
                check=1;
              }
              break;
    }
  } while (!feof(f));
  free(buf);
  fclose(f);
  return check;
}

int pass3(void) {
  FILE * f;
  char * buf, * dosfname, * cflfname,* fbuf;
  int i,a;
  struct resinfostruct * res;
  int nameofs=0,strbufofs;
  strbufofs=*(int*)outbuf+sizeof(struct resinfostruct)*resources;
  res=(void*)(outbuf+*(int*)outbuf);
  dosfname=malloc(256);
  cflfname=malloc(256);
  buf=malloc(256);
  f=fopen("cfl2.ini","rb");
  while (!feof(f)) {
    memset(dosfname,0,255);
    memset(cflfname,0,255);
    read_line(f,buf);
    switch (*buf) {
    case ';':
    case  0 :
    case '#': break;
    default : i=0;
              while ((*(buf+i)!='=') && (*(buf+i)!=0)) {
                *(dosfname+i)=*(buf+i);
                i++;
              }
              *(dosfname+i)=0;
              printf("\t\"%s\"\n",dosfname);
              a=0; i++;
              while ((*(buf+i)!=',') && (*(buf+i)!=0)) {
                *(cflfname+a)=*(buf+i);
                i++; a++;
              }
              *(cflfname+a)=0;
              if (strcmp(cflfname,"*")==0)
                memcpy(cflfname,dosfname,255);
              _strupr(cflfname);
              printf("\t  ->\"%s\"\n",cflfname);
              res->algo=strtol(buf+1+i,0,0);
              res->nameofs=nameofs;
              i=0;
              while (*(cflfname+i)) {
                *(outbuf+strbufofs+nameofs)=*(cflfname+i);
                i++;
                nameofs++;
              }
              nameofs++;
              res->ofs=ftell(outf);
              res->decompressed_len=filelen(dosfname);
              fbuf=getfile(dosfname);
              printf("\t    Algo: %8d Length: %8d\n",res->algo,res->decompressed_len);
              fbuf=CFL_pack(fbuf,res->decompressed_len,&res->len,res->algo,&res->csum);
              printf("\t    Csum: %08X Packed: %8d (%3.6f%%)\n",res->csum,res->len,(res->len*100.0)/res->decompressed_len);
              if (fbuf==NULL) exit(1);
              fwrite(fbuf,res->len,1,outf);
              res++;
              break;
    }
  }
  free(buf);
  free(dosfname);
  free(cflfname);
  fclose(f);
  return strbufofs+nameofs;
}

void main(void) {
  int i;
  int hdrofs;
  struct cflheaderstruct hdr;
  setbuf(stdout,NULL);
  if (!CFL_dosexists("cfl2.ini")) {
    printf("\n\nMakelib only runs by reading cfl2.ini and processing"
           "\nthe files and config options mentioned there.\n");
    return;
  }
  printf("Makelib for CFL 2.0\n");
  printf("Step 1: reading configuration\n");
  ini_pass1();
  printf("Configuration:"
         "\n\tTarget file : \"%s\""
         "\n\tInfo string : \"%s\""
         "\n\tCheck string: \"%s\""
         "\n\tLib compress: %d"
         "\n\tFiles       : %d\n",
         (targetfile==NULL)?"-NOT DEFINED-":targetfile,
         infostring,
         checkstring,
         libpack,
         resources);
  if (targetfile==NULL) {
    printf("\n\nTarget file must be defined.\n");
    return;
  }
  if (!resources) {
    printf("\n\nNo data files.\n");
    return;
  }
  if (!(libpack%100)) {
    printf("\n\nLibrary must be compressed.\n");
    return;
  }
  outf=fopen(targetfile,"wb");
  i=HEADERMAGIC;
  fwrite((char*)&i,4,1,outf);
  i=strlen(infostring);
  fwrite((char*)&i,4,1,outf);
  i=0;
  while (*(infostring+i)) {
    fputc(*(infostring+i),outf);
    i++;
  }
  hdrofs=ftell(outf);
  fwrite((char*)&hdr,sizeof(struct cflheaderstruct),1,outf);
  outbuf=malloc((sizeof(struct resinfostruct)*resources)+resources*256);
  memset(outbuf,0,(sizeof(struct resinfostruct)*resources)+resources*256);
  *((int*)outbuf)=4*3+strlen(checkstring);
  *((int*)outbuf+1)=resources;
  *((int*)outbuf+2)=strlen(checkstring);
  i=0;
  while (*(checkstring+i)) {
    *(outbuf+4*3+i)=*(checkstring+i);
    i++;
  }
  printf("Step 2: Checking files\n");
  if (pass2()) return;
  printf("Step 3: Building library\n");
  hdr.decompressed_len=pass3();
  printf("Step 4: Writing library\n");
  hdr.algo=libpack;
  printf("\t  ->\"CFL 2.0 library info block\"\n");
  printf("\t    Algo: %8d Length: %8d\n",hdr.algo,hdr.decompressed_len);
  outbuf=CFL_pack(outbuf,hdr.decompressed_len,&hdr.len,hdr.algo,&hdr.csum);
  printf("\t    Csum: %08X Packed: %8d (%3.6f%%)\n",hdr.csum,hdr.len,(hdr.len*100.0)/hdr.decompressed_len);
  hdr.ofs=ftell(outf);
  fwrite(outbuf,hdr.len,1,outf);
  printf("Step 5: Writing footer and header\n");
  i=ftell(outf)+8;
  fwrite((char*)&i,4,1,outf);
  i=FOOTERMAGIC;
  fwrite((char*)&i,4,1,outf);
  fseek(outf,hdrofs,SEEK_SET);
  fwrite((char*)&hdr,sizeof(struct cflheaderstruct),1,outf);
  printf("Done.\n");
  fclose(outf);
}
